
module Kameleoon
  NONCE_LENGTH = 16

  module DataType
    CUSTOM = "CUSTOM"
    BROWSER = "BROWSER"
    CONVERSION = "CONVERSION"
    INTEREST = "INTEREST"
    PAGE_VIEW = "PAGE_VIEW"
  end

  module BrowserType
    CHROME = 0
    INTERNET_EXPLORER = 1
    FIREFOX = 2
    SAFARI = 3
    OPERA = 4
    OTHER = 5
  end

  class Data
    attr_accessor :instance, :sent

    def initialize(*args)
      raise KameleoonError.new("Cannot initialise interface.")
    end

    def obtain_full_post_text_line
      raise KameleoonError.new("ToDo: implement this method.")
    end
  end

  class CustomData < Data
    attr_accessor :id, :value

    # @param [Integer] id Id of the custom data
    # @param [String] value Value of the custom data
    #
    # @overload
    # @param [Hash] hash Json value encoded in a hash.
    def initialize(*args)
      @instance = DataType::CUSTOM
      @sent = false
      unless args.empty?
        if args.length == 1
          hash = args.first
          if hash["id"].nil?
            raise NotFoundError.new("id")
          end
          @id = hash["id"].to_s
          if hash["value"].nil?
            raise NotFoundError.new(hash["value"])
          end
          @value = hash["value"]
        elsif args.length == 2
          @id = args[0].to_s
          @value = args[1]
        end
      end
    end

    def obtain_full_post_text_line
      to_encode = "[[\"" + @value.to_s.gsub("\"", "\\\"") + "\",1]]"
      nonce = Kameleoon::Utils.generate_random_string(NONCE_LENGTH)
      "eventType=customData&index=" + @id.to_s + "&valueToCount=" + URI.encode_www_form_component(to_encode) + "&overwrite=true&nonce=" + nonce
    end
  end

  class Browser < Data
    attr_accessor :browser

    # @param [BrowserType] browser_type Browser type, can be: CHROME, INTERNET_EXPLORER, FIREFOX, SAFARI, OPERA, OTHER
    def initialize(browser_type)
      @instance = DataType::BROWSER
      @sent = false
      @browser = browser_type
    end

    def obtain_full_post_text_line
      nonce = Kameleoon::Utils.generate_random_string(NONCE_LENGTH)
      "eventType=staticData&browser=" + @browser.to_s + "&nonce=" + nonce
    end
  end

  class PageView < Data
    attr_accessor :url, :title, :referrer

    # @param [String] url Url of the page
    # @param [String] title Title of the page
    # @param [Integer] referrer Optional field - Referrer id
    def initialize(url, title, referrer = nil)
      @instance = DataType::PAGE_VIEW
      @sent = false
      @url = url
      @title = title
      @referrer = referrer
    end

    def obtain_full_post_text_line
      nonce = Kameleoon::Utils.generate_random_string(NONCE_LENGTH)
      referrer_text = ""
      unless @referrer.nil?
        referrer_text = "&referrers=[" + @referrer + "]"
      end
      "eventType=page&href=" + @url + "&title=" + @title + "&keyPages=[]" + referrer_text + "&nonce=" + nonce
    end
  end

  class Conversion < Data
    attr_accessor :goal_id, :revenue, :negative

    # @param [Integer] goal_id Id of the goal associated to the conversion
    # @param [Float] revenue Optional field - Revenue associated to the conversion.
    # @param [Boolean] negative Optional field - If the revenue is negative. By default it's positive.
    def initialize(goal_id, revenue = 0.0, negative = false)
      @instance = DataType::CONVERSION
      @sent = false
      @goal_id = goal_id
      @revenue = revenue
      @negative = negative
    end

    def obtain_full_post_text_line
      nonce = Kameleoon::Utils.generate_random_string(NONCE_LENGTH)
      "eventType=conversion&goalId=" + @goal_id.to_s + "&revenue=" + @revenue.to_s + "&negative=" + @negative.to_s + "&nonce=" + nonce
    end
  end

  class Interest < Data
    attr_accessor :index

    # @param [Integer] index Index of the interest
    def initialize(index)
      @instance = DataType::INTEREST
      @sent = false
      @index = index
    end

    def obtain_full_post_text_line
      nonce = Kameleoon::Utils.generate_random_string(NONCE_LENGTH)
      "eventType=interests&indexes=[" + @index.to_s + "]&fresh=true&nonce=" + nonce
    end
  end
end